"""
Convert a Open Street Maps `.map` format file into a networkx directional graph.

This parser is based on the osm to networkx tool
from Loïc Messal (github : Tofull)

Added :
- python3.6 compatibility
- networkx v2 compatibility
- cache to avoid downloading the same osm tiles again and again
- distance computation to estimate length of each ways (useful to compute the shortest path)

"""

import copy
import xml.sax
import logging
from collections import namedtuple

import networkx as nx
import requests

from evrouting.graph_tools import CHARGING_COEFFICIENT_KEY
from evrouting.osm.const import ms_to_kmh
from evrouting.osm.profiles import speed
from evrouting.osm.routing import find_nearest, haversine_distance

logger = logging.getLogger(__name__)

OsrmConf = namedtuple('OsrmConf', ['server', 'port', 'version', 'profile'],
                      defaults=('v1', 'driving'))


def query_url(service, coordinates, osrm_config: OsrmConf):
    """Construct query url."""
    return f'http://{osrm_config.server}:{osrm_config.port}' \
           f'/{service}/{osrm_config.version}/{osrm_config.profile}/' \
           f'{";".join([f"{lon},{lat}" for lat, lon in coordinates])}'


def insert_charging_stations(G, charging_stations):
    """Insert Charging Stations"""
    S = set()
    for s in charging_stations:
        lon = s['lon']
        lat = s['lat']
        n = find_nearest(G, (lat, lon), distance_limit=500)
        if n:
            G.nodes[n][CHARGING_COEFFICIENT_KEY] = s['power']
            S.add(n)

    return S


class OsrmDistance:

    def __init__(self, osrm_config: OsrmConf = None):
        self.osrm_config = osrm_config or OsrmConf(
            server='localhost',
            port=5000
        )

    def __call__(self, G, u, v):
        """Calc distance between u and v on osrm."""
        coordinates = [
            (G.nodes[u]['lat'], G.nodes[u]['lon']),
            (G.nodes[v]['lat'], G.nodes[v]['lon'])
        ]
        url = query_url('route', coordinates, self.osrm_config)

        resp = requests.get(url, timeout=0.1)
        resp.raise_for_status()
        resp = resp.json()
        return resp['routes'][0]['duration']


def read_osm(osm_xml_data, profile) -> nx.DiGraph:
    """
    Read graph in OSM format from file specified by name or by stream object.
    Create Graph containing all highways as edges.

    """

    osm = OSM(osm_xml_data)
    G = nx.DiGraph()

    # Add ways
    for w in osm.ways.values():
        if 'highway' not in w.tags:
            continue
        if w.tags['highway'] not in profile['highway_whitelist']:
            continue

        for u_id, v_id in zip(w.nds[:-1], w.nds[1:]):
            u, v = osm.nodes[u_id], osm.nodes[v_id]

            # Travel-time from u to v
            d = haversine_distance(
                u.lon, u.lat, v.lon, v.lat, unit_m=True
            ) / speed(w, profile) * ms_to_kmh

            if w.tags.get('oneway', 'no') == 'yes':
                # ONLY ONE DIRECTION
                G.add_edge(u_id, v_id, weight=d)
            else:
                # BOTH DIRECTION
                G.add_edge(u_id, v_id, weight=d)
                G.add_edge(v_id, u_id, weight=d)

    # Complete the used nodes' information
    for n_id in G.nodes():
        n = osm.nodes[n_id]
        G.nodes[n_id]['lat'] = n.lat
        G.nodes[n_id]['lon'] = n.lon
        G.nodes[n_id]['id'] = n.id

    return G


class Node(object):
    def __init__(self, id, lon, lat):
        self.id = id
        self.lon = lon
        self.lat = lat
        self.tags = {}

    def __str__(self):
        return "Node (id : %s) lon : %s, lat : %s " % (self.id, self.lon, self.lat)


class Way(object):
    def __init__(self, id):
        self.id = id
        self.nds = []
        self.tags = {}

    def split(self, node_pass_count):
        """
        Slice way at every crossing i.e. when a waypoint is passend by
        multiple ways.
        """

        # slice the node-array using this nifty recursive function
        def slice_array(waypoints):
            for i in range(1, len(waypoints) - 1):
                if node_pass_count[waypoints[i]] > 1:
                    left = waypoints[:i + 1]
                    right = waypoints[i:]

                    rightsliced = slice_array(right)

                    return [left] + rightsliced
            return [waypoints]

        slices = slice_array(self.nds)

        # create a way object for each node-array slice
        ret = []
        i = 0
        for slice in slices:
            littleway = copy.copy(self)
            littleway.id += "-%d" % i
            littleway.nds = slice
            ret.append(littleway)
            i += 1

        return ret


class OSM(object):
    def __init__(self, osm_xml_data):
        """ File can be either a filename or stream/file object.

        set `is_xml_string=False` if osm_xml_data is a filename or a file stream.
        """
        nodes = {}
        ways = {}

        class OSMHandler(xml.sax.ContentHandler):
            @classmethod
            def setDocumentLocator(self, loc):
                pass

            @classmethod
            def startDocument(self):
                pass

            @classmethod
            def endDocument(self):
                pass

            @classmethod
            def startElement(self, name, attrs):
                if name == 'node':
                    self.currElem = Node(attrs['id'], float(attrs['lon']), float(attrs['lat']))
                elif name == 'way':
                    self.currElem = Way(attrs['id'])
                elif name == 'tag':
                    self.currElem.tags[attrs['k']] = attrs['v']
                elif name == 'nd':
                    self.currElem.nds.append(attrs['ref'])

            @classmethod
            def endElement(self, name):
                if name == 'node':
                    nodes[self.currElem.id] = self.currElem
                elif name == 'way':
                    ways[self.currElem.id] = self.currElem

            @classmethod
            def characters(self, chars):
                pass

        xml.sax.parse(osm_xml_data, OSMHandler)

        self.nodes = nodes
        self.ways = ways

        # count times each node is used
        node_histogram = dict.fromkeys(self.nodes.keys(), 0)
        for way in self.ways.values():
            if len(way.nds) < 2:  # if a way has only one node, delete it out of the osm collection
                del self.ways[way.id]
            else:
                for node in way.nds:
                    node_histogram[node] += 1

        # use that histogram to split all ways, replacing the member set of ways
        new_ways = {}
        for id, way in self.ways.items():
            split_ways = way.split(node_histogram)
            for split_way in split_ways:
                new_ways[split_way.id] = split_way
        self.ways = new_ways
