from typing import Tuple
from math import radians, cos, sin, asin, sqrt

import networkx as nx

from evrouting.osm.const import ms_to_kmh

lat = float
lon = float
point = Tuple[lat, lon]


def haversine_distance(lon1, lat1, lon2, lat2, unit_m=True):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    default unit : km
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))
    r = 6371  # Radius of the Earth in kilometers. Use 3956 for miles
    if unit_m:
        r *= 1000
    return c * r


def find_nearest(G, v: point, distance_limit=None):
    """
    Find nearest point to location v within radius
    of distance_limit.
    """
    min_dist = None
    closest_node = None

    lat_v, lon_v = v

    for n in G.nodes:
        d = haversine_distance(
            G.nodes[n]['lat'],
            G.nodes[n]['lon'],
            lat_v,
            lon_v,
            unit_m=True
        )
        if min_dist is None or d < min_dist:
            closest_node = n
            min_dist = d

    if distance_limit:
        return closest_node if min_dist <= distance_limit else None
    else:
        return closest_node


def shortest_path(G, s: point, t: point, profile):
    """Calc A* shortest path."""
    _s = find_nearest(G, s)
    _t = find_nearest(G, t)

    def dist(u, v):
        return haversine_distance(
            G.nodes[u]['lat'],
            G.nodes[u]['lon'],
            G.nodes[v]['lat'],
            G.nodes[v]['lon'],
            unit_m=True
        ) / profile['maxspeed'] * ms_to_kmh

    return nx.astar_path(G, _s, _t, heuristic=dist)


def to_coordinates(G, path):
    """
    Path of nodes to path of coordinates.

    Note: Coordinates are (lon, lat) to conform to
        geojson.
    """
    def get_coordinates(n):
        lat = G.nodes[n]['lat']
        lon = G.nodes[n]['lon']
        return lon, lat

    return list(map(get_coordinates, path))
